@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009, 2010, 2011
+ * Copyright (c) 2009, 2010, 2011, 2012, 2013
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,24 +44,24 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: iobuf.c,v 1.80 2011/07/21 13:48:41 michaels Exp $";
+"$Id: iobuf.c,v 1.116 2013/10/25 12:55:01 karls Exp $";
 
 static int socks_flushallbuffers(void);
 /*
- * Flushes all buffers.  
- * Returns 0 if all were flushed successfully. 
- * Returns -1 if we failed to completly flush at least one buffer.
+ * Flushes all buffers.
+ * Returns 0 if all were flushed successfully.
+ * Returns -1 if we failed to completely flush at least one buffer.
  */
 
 #if !SOCKS_CLIENT
 /*
- * - Each negotiatechild client can use one iobuffer for control.
- * - Each requestchild and iochild client can use one iobuffer for
+ * - Each negotiate child client can use one iobuffer for control.
+ * - Each request child and io child client can use one iobuffer for
  *   control, src, and st.
  */
 static iobuffer_t
    iobufv[MAX((SOCKD_NEGOTIATEMAX * 1  /* control */),
-               MAX((SOCKD_IOMAX   * 3  /* control, src, dst */),
+               MAX((SOCKD_IOMAX   * 4  /* control, src, dst1, dst2. */),
                     (SOCKD_REQUESTMAX   * 3  /* control, src, dst */)))];
 static const size_t  iobufc = ELEMENTS(iobufv);
 
@@ -73,46 +73,70 @@ static size_t     iobufc;
 static size_t lastfreei;  /* last buffer freed, for quick allocation.  */
 
 void
-socks_setbuffer(s, mode, size)
-   const int s;
+socks_setbuffer(iobuf, mode, size)
+   iobuffer_t *iobuf;
    const int mode;
    ssize_t size;
 {
-   iobuffer_t *iobuf;
-
-   SASSERTX(size <= SOCKD_BUFSIZE); 
-   if ((iobuf = socks_getbuffer(s)) == NULL)
-      return;
 
    iobuf->info[READ_BUF].mode  = _IONBF; /* only one supported for read. */
    iobuf->info[WRITE_BUF].mode = mode;
 
    if (size == -1)
       size = sizeof(*iobuf->buf);
-   else
-      SASSERTX(size <= (ssize_t)sizeof(*iobuf->buf));
+
+   SASSERTX(size > 0);
+   SASSERTX(size <= (ssize_t)sizeof(*iobuf->buf));
 
    iobuf->info[READ_BUF].size  = size;
    iobuf->info[WRITE_BUF].size = size;
 }
 
+void
+socks_setbufferfd(s, mode, size)
+   const int s;
+   const int mode;
+   ssize_t size;
+{
+   iobuffer_t *iobuf;
+
+   SASSERTX(size <= SOCKD_BUFSIZE);
+
+   if ((iobuf = socks_getbuffer(s)) == NULL)
+      return;
+
+   socks_setbuffer(iobuf, mode, size);
+}
+
+
 int
-socks_flushbuffer(s, len)
+socks_flushbuffer(s, len, sendtoflags)
    const int s;
    const ssize_t len;
+   sendto_info_t *sendtoflags;
 {
    const char *function = "socks_flushbuffer()";
-   ssize_t written, encoded, towrite;
-   unsigned char buf[SOCKD_BUFSIZE];
+   size_t towrite;
+   ssize_t written;
+   unsigned char inputmem[sizeof(iobuffer_t)];
+   int encoded;
 
    if (sockscf.option.debug >= DEBUG_VERBOSE)
-      slog(LOG_DEBUG, "%s: socket %d, len = %ld", function, s, (long)len);
+      slog(LOG_DEBUG, "%s: fd %d, len = %ld", function, s, (long)len);
+
+   if (sendtoflags != NULL)
+      sendtoflags->tosocket = 0;
 
    if (s == -1)
       return socks_flushallbuffers();
 
-   if (socks_bufferhasbytes(s, WRITE_BUF) == 0)
+   if (!socks_bufferhasbytes(s, WRITE_BUF))
       return 0;
+   else
+      slog(LOG_DEBUG, "%s: buffer for fd %d has bytes (%lu).  Flushing",
+           function,
+           s,
+           (unsigned long)socks_bytesinbuffer(s, WRITE_BUF, 1));
 
 #if SOCKS_CLIENT && HAVE_GSSAPI
    /*
@@ -121,7 +145,7 @@ socks_flushbuffer(s, len)
     * end up creating 100 gssapi-encoded one-byte packets.
     * We therefore postpone encoding til we get a flush call.
     *
-    * Note that we only use the buffer in the client if we are doing
+    * Note that we only use the iobuffer in the client if we are doing
     * gssapi i/o, never for ordinary i/o.  For the server, it is used
     * for both ordinary and gssapi-based i/o.
     */
@@ -133,29 +157,51 @@ socks_flushbuffer(s, len)
        * Already have encoded data ready for write.  Must always
        * write that first, since it came first.
        */
-      struct socksfd_t socksfd, *p;
+      socksfd_t socksfd, *p;
 
       p = socks_getaddr(s, &socksfd, 1);
       SASSERTX(p != NULL);
       SASSERTX(socksfd.state.auth.method == AUTHMETHOD_GSSAPI);
 
-      towrite = socks_getfrombuffer(s, WRITE_BUF, 1, buf, sizeof(buf));
+      towrite = socks_getfrombuffer(s,
+                                    0,
+                                    WRITE_BUF,
+                                    1,
+                                    inputmem,
+                                    sizeof(inputmem));
 
       if (sockscf.option.debug >= DEBUG_VERBOSE)
          slog(LOG_DEBUG, "%s: flushing %lu encoded byte%s ...",
-         function, (long unsigned)towrite, towrite == 1 ? "" : "s");
+              function, (long unsigned)towrite, towrite == 1 ? "" : "s");
+
+      /*
+       * this is important since it verifies that we fetched all
+       * the data from the buffer, so that what we add now does
+       * not erroneously get appended after something, since when
+       * we fetched it was first.
+       */
 
       SASSERTX(socks_bytesinbuffer(s, WRITE_BUF, 1) == 0);
 
-      if ((written = socks_sendton(s, buf, towrite, towrite, 0, NULL, 0, NULL))
-      != towrite) {
+      written = sendto(s, inputmem, towrite, 0, NULL, 0);
+
+      if (written != -1 && sendtoflags != NULL)
+         sendtoflags->tosocket += written;
+
+      if (written == -1 || (size_t)written != towrite) {
          if (sockscf.option.debug >= DEBUG_VERBOSE)
-            slog(LOG_DEBUG, "%s: socks_sendton() flushed only %ld/%lu: %s",
-            function, (long)written, (long unsigned)towrite, strerror(errno));
+            slog(LOG_DEBUG, "%s: sendton() flushed only %ld/%lu: %s",
+                 function,
+                 (long)written,
+                 (long unsigned)towrite,
+                 strerror(errno));
 
          if (written > 0) { /* add back what we failed to write. */
-            socks_addtobuffer(s, WRITE_BUF, 1, &buf[written],
-            towrite - written);
+            socks_addtobuffer(s,
+                              WRITE_BUF,
+                              1,
+                              &inputmem[written],
+                              towrite - written);
 
             continue;
          }
@@ -172,42 +218,58 @@ socks_flushbuffer(s, len)
       /*
        * Unencoded data in buffer, need to encode it first.
        */
-      struct socksfd_t socksfd, *p;
-      unsigned char token[GSSAPI_HLEN + MAXGSSAPITOKENLEN];
-      unsigned short tokenlen;
-      size_t token_length;
+      gss_buffer_desc input_token, output_token;
+      socksfd_t socksfd, *ptr;
+      unsigned char outputmem[GSSAPI_HLEN + MAXGSSAPITOKENLEN];
+      unsigned short pshort;
       ssize_t toencode;
 
-      p = socks_getaddr(s, &socksfd, 1);
-      SASSERTX(p != NULL);
+      ptr = socks_getaddr(s, &socksfd, 1);
+      SASSERTX(ptr != NULL);
       SASSERTX(socksfd.state.auth.method == AUTHMETHOD_GSSAPI);
 
-      toencode = socks_getfrombuffer(s,
-                                     WRITE_BUF,
-                                     0,
-                                     buf,
-                                     MIN(sizeof(buf),
+      toencode
+      = socks_getfrombuffer(s,
+                            0,
+                            WRITE_BUF,
+                            0,
+                            inputmem,
+                            MIN(sizeof(inputmem),
                              socksfd.state.auth.mdata.gssapi.state.maxgssdata));
 
       if (sockscf.option.debug >= DEBUG_VERBOSE)
          slog(LOG_DEBUG, "%s: encoding %ld byte%s before flushing ...",
-         function, (long)toencode, toencode == 1 ? "" : "s");
+              function, (long)toencode, toencode == 1 ? "" : "s");
 
-      token_length = sizeof(token) - GSSAPI_HLEN;
-      if (gssapi_encode(buf, toencode, &socksfd.state.auth.mdata.gssapi.state,
-      token + GSSAPI_HLEN, &token_length) != 0) {
+      input_token.value  = inputmem;
+      input_token.length = toencode;
+
+      output_token.value  = outputmem + GSSAPI_HLEN;
+      output_token.length = sizeof(outputmem) - GSSAPI_HLEN;
+
+      if (gssapi_encode(&input_token,
+                        &socksfd.state.auth.mdata.gssapi.state,
+                        &output_token) != 0) {
          swarnx("%s: gssapi_encode() failed", function);
          return -1;
       }
 
-      token[GSSAPI_VERSION] = SOCKS_GSSAPI_VERSION;
-      token[GSSAPI_STATUS]  = SOCKS_GSSAPI_PACKET;
+      ((char *)(output_token.value))[GSSAPI_VERSION] = SOCKS_GSSAPI_VERSION;
+      ((char *)(output_token.value))[GSSAPI_STATUS]  = SOCKS_GSSAPI_PACKET;
 
-      tokenlen = htons(token_length);
-      memcpy(&token[GSSAPI_TOKEN_LENGTH], &tokenlen, sizeof(tokenlen));
+      pshort = htons(output_token.length);
+      memcpy(&((char *)output_token.value)[GSSAPI_TOKEN_LENGTH],
+             &pshort,
+             sizeof(pshort));
 
-      SASSERTX(GSSAPI_HLEN + token_length <= socks_freeinbuffer(s, WRITE_BUF));
-      socks_addtobuffer(s, WRITE_BUF, 1, token, GSSAPI_HLEN + token_length);
+      SASSERTX(GSSAPI_HLEN + output_token.length
+      <=       socks_freeinbuffer(s, WRITE_BUF));
+
+      socks_addtobuffer(s,
+                        WRITE_BUF,
+                        1,
+                        output_token.value,
+                        output_token.length + GSSAPI_HLEN);
    }
 
    if (socks_bufferhasbytes(s, WRITE_BUF) == 0)
@@ -216,11 +278,6 @@ socks_flushbuffer(s, len)
 
    written = 0;
    do {
-      /*
-       * In case of client, we want to keep trying unless error is permanent.
-       * In case of server, we want to keep writing while we can, but return
-       * at the first error, permanent or not.
-       */
       ssize_t rc;
 
       if (socks_bytesinbuffer(s, WRITE_BUF, 0) > 0) {
@@ -234,23 +291,34 @@ socks_flushbuffer(s, len)
       else
          SERRX(0);
 
+      /*
+       * In case of client, we want to keep trying unless error is permanent.
+       * In case of server, we want to keep writing while we can, but return
+       * at the first error, permanent or not.
+       */
       towrite = socks_getfrombuffer(s,
+                                    0,
                                     WRITE_BUF,
                                     encoded,
-                                    buf, len == -1 ? sizeof(buf) : (size_t)len);
+                                    inputmem,
+                                    len == -1 ? sizeof(inputmem) : (size_t)len);
 
-      rc = sendto(s, buf, towrite, 0, NULL, 0);
+      rc = sendto(s, inputmem, towrite, 0, NULL, 0);
 
-      if (sockscf.option.debug >= DEBUG_VERBOSE)
-         slog(LOG_DEBUG, "%s: flushed %ld/%ld %s byte%s (%s), 0x%x, 0x%x",
-         function, (long)rc, (long)towrite, encoded ? "encoded" : "unencoded",
-         rc == 1 ? "" : "s",
-         strerror(errno),
-         buf[rc - 2], buf[rc - 1]);
+      if (rc != -1 && sendtoflags != NULL)
+         sendtoflags->tosocket += rc;
+
+      slog(LOG_DEBUG, "%s: flushed %ld/%ld %s byte%s (%s)",
+           function,
+           (long)rc,
+           (long)towrite,
+           encoded ? "encoded" : "unencoded",
+           rc == 1 ? "" : "s",
+           strerror(errno));
 
       if (rc == -1) {
          /* could not flush anything, add all back. */
-         socks_addtobuffer(s, WRITE_BUF, encoded, buf, towrite);
+         socks_addtobuffer(s, WRITE_BUF, encoded, inputmem, towrite);
 
 #if SOCKS_CLIENT
          if (ERRNOISTMP(errno)) {
@@ -275,7 +343,11 @@ socks_flushbuffer(s, len)
       }
 
       written += rc;
-      socks_addtobuffer(s, WRITE_BUF, encoded, buf + rc, towrite - rc);
+      socks_addtobuffer(s,
+                        WRITE_BUF,
+                        encoded,
+                        inputmem + rc,
+                        towrite - (size_t)rc);
 
    } while ((len == -1 || written < len)
    && socks_bytesinbuffer(s, WRITE_BUF, encoded) > 0);
@@ -326,13 +398,14 @@ socks_allocbuffer(s, stype)
    const int stype;
 {
    const char *function = "socks_allocbuffer()";
+
 #if SOCKS_CLIENT
    sigset_t oset;
 #endif /* SOCKS_CLIENT */
    iobuffer_t *freebuffer;
    size_t i;
 
-   slog(LOG_DEBUG, "%s: socket %d, stype = %d", function, s, stype);
+   slog(LOG_DEBUG, "%s: fd %d, stype = %d", function, s, stype);
 
    SASSERTX(socks_getbuffer(s) == NULL);
 
@@ -341,12 +414,13 @@ socks_allocbuffer(s, stype)
     */
    if (lastfreei < iobufc && !iobufv[lastfreei].allocated)
       freebuffer = &iobufv[lastfreei];
-   else
+   else {
       for (i = 0, freebuffer = NULL; i < iobufc; ++i)
          if (!iobufv[i].allocated) {
             freebuffer = &iobufv[i];
             break;
          }
+   }
 
 #if SOCKS_CLIENT
    /*
@@ -358,26 +432,24 @@ socks_allocbuffer(s, stype)
    socks_sigblock(SIGIO, &oset);
 
    if (freebuffer == NULL) {
-      if ((iobufv = realloc(iobufv, sizeof(*iobufv) * ++iobufc))
-      == NULL) {
+      void *p;
+
+      if ((p = realloc(iobufv, sizeof(*iobufv) * (iobufc + 1))) == NULL) {
          swarn("%s: %s", function, NOMEM);
          socks_sigunblock(&oset);
 
          return NULL;
       }
 
+      iobufv     = p;
+      iobufc     += 1;
+
       freebuffer = &iobufv[iobufc - 1];
    }
 #endif /* SOCKS_CLIENT */
 
    SASSERTX(freebuffer != NULL);
-
-   bzero(freebuffer, sizeof(*freebuffer));
-   freebuffer->s         = s;
-   freebuffer->stype     = stype;
-   freebuffer->allocated = 1;
-
-   socks_setbuffer(s, _IONBF, -1); /* default; no buffering. */
+   socks_initbuffer(s, stype, freebuffer);
 
 #if SOCKS_CLIENT
    socks_sigunblock(&oset);
@@ -386,6 +458,22 @@ socks_allocbuffer(s, stype)
    return freebuffer;
 }
 
+void
+socks_initbuffer(fd, stype, iobuf)
+   const int fd;
+   const int stype;
+   iobuffer_t *iobuf;
+{
+
+   bzero(iobuf, sizeof(*iobuf));
+   iobuf->s         = fd;
+   iobuf->stype     = stype;
+   iobuf->allocated = 1;
+
+   socks_setbuffer(iobuf, _IONBF, -1); /* default; no buffering. */
+}
+
+
 void
 socks_reallocbuffer(old, new)
    const int old;
@@ -395,7 +483,7 @@ socks_reallocbuffer(old, new)
    iobuffer_t *iobuf = socks_getbuffer(old);
 
    slog(LOG_DEBUG, "%s: old %d, new %d, %s",
-   function, old, new, iobuf == NULL ? "no iobuf" : "have iobuf");
+        function, old, new, iobuf == NULL ? "no iobuf" : "have iobuf");
 
    if (iobuf != NULL)
       iobuf->s = new;
@@ -407,31 +495,32 @@ socks_freebuffer(s)
 {
    const char *function = "socks_freebuffer()";
 
-   slog(LOG_DEBUG, "%s: socket %d", function, s);
+   slog(LOG_DEBUG, "%s: fd %d", function, s);
 
    if (lastfreei < iobufc
    && iobufv[lastfreei].s == s && iobufv[lastfreei].allocated)
-      /* LINTED */ /* EMPTY */;
+      ;
    else
       lastfreei = 0;
 
-   for (; lastfreei < iobufc; ++lastfreei)
-      if ( iobufv[lastfreei].s == s && iobufv[lastfreei].allocated) {
-         slog(LOG_DEBUG, "%s: freeing buffer %d", function, s);
+   for (; lastfreei < iobufc; ++lastfreei) {
+      if (!iobufv[lastfreei].allocated
+      ||  iobufv[lastfreei].s != s)
+         continue;
 
-         if (sockscf.option.debug >= DEBUG_VERBOSE
-         && ( socks_bufferhasbytes(s, READ_BUF)
-           || socks_bufferhasbytes(s, WRITE_BUF)))
-            slog(LOG_DEBUG, "%s: freeing buffer with data (%lu/%lu, %lu/%lu)",
-            function,
-            (unsigned long)socks_bytesinbuffer(s, READ_BUF, 0),
-            (unsigned long)socks_bytesinbuffer(s, READ_BUF, 1),
-            (unsigned long)socks_bytesinbuffer(s, WRITE_BUF, 0),
-            (unsigned long)socks_bytesinbuffer(s, WRITE_BUF, 1));
+      if (sockscf.option.debug >= DEBUG_VERBOSE
+      && ( socks_bufferhasbytes(s, READ_BUF)
+        || socks_bufferhasbytes(s, WRITE_BUF)))
+         slog(LOG_DEBUG, "%s: freeing buffer with data (%lu/%lu, %lu/%lu)",
+         function,
+         (unsigned long)socks_bytesinbuffer(s, READ_BUF, 0),
+         (unsigned long)socks_bytesinbuffer(s, READ_BUF, 1),
+         (unsigned long)socks_bytesinbuffer(s, WRITE_BUF, 0),
+         (unsigned long)socks_bytesinbuffer(s, WRITE_BUF, 1));
 
-         iobufv[lastfreei].allocated = 0;
-         return;
-      }
+      iobufv[lastfreei].allocated = 0;
+      return;
+   }
 }
 
 size_t
@@ -452,19 +541,28 @@ socks_addtobuffer(s, which, encoded, data, datalen)
    iobuf = socks_getbuffer(s);
    SASSERTX(iobuf != NULL);
 
+   if (iobuf->stype == SOCK_DGRAM) { /* no buffering of udp for now. */
+      SASSERTX(socks_bufferhasbytes(s, READ_BUF)  == 0);
+      SASSERTX(socks_bufferhasbytes(s, WRITE_BUF) == 0);
+
+      SERRX(0);
+   }
+
    toadd = MIN(socks_freeinbuffer(s, which), datalen);
 
    if (sockscf.option.debug >= DEBUG_VERBOSE)
-      slog(LOG_DEBUG, "%s: s = %lu, add %lu %s byte%s to %s buffer that "
-                      "currently has %lu decoded, %lu encoded",
-                      function,
-                      (long unsigned)s,
-                      (unsigned long)datalen,
-                      encoded ? "encoded" : "decoded",
-                      datalen == 1 ? "" : "s",
-                      which == READ_BUF ? "read" : "write",
-                      (unsigned long)socks_bytesinbuffer(s, which, 0),
-                      (unsigned long)socks_bytesinbuffer(s, which, 1));
+      slog(LOG_DEBUG,
+           "%s: fd = %d, add %lu %s byte%s to %s buffer which currently has "
+           "%lu decoded, %lu encoded.  Last byte to add: 0x%x",
+           function,
+           s,
+           (unsigned long)datalen,
+           encoded ? "encoded" : "decoded",
+           datalen == 1 ? "" : "s",
+           which == READ_BUF ? "read" : "write",
+           (unsigned long)socks_bytesinbuffer(s, which, 0),
+           (unsigned long)socks_bytesinbuffer(s, which, 1),
+           (int)((const unsigned char *)data)[datalen - 1]);
 
       SASSERTX(toadd >= datalen);
 
@@ -474,7 +572,9 @@ socks_addtobuffer(s, which, encoded, data, datalen)
           * the end of the buffer.
           */
          memcpy(&iobuf->buf[which][socks_bytesinbuffer(s, which, 0)
-         + socks_bytesinbuffer(s, which, 1)], data, toadd);
+                                 + socks_bytesinbuffer(s, which, 1)],
+                data,
+                toadd);
 
          iobuf->info[which].enclen += toadd;
       }
@@ -508,14 +608,18 @@ socks_bytesinbuffer(s, which, encoded)
    const int encoded;
 {
    const iobuffer_t *iobuf = socks_getbuffer(s);
+   size_t rc;
 
    if (iobuf == NULL)
       return 0;
 
    if (encoded)
-      return iobuf->info[which].enclen;
+      rc = iobuf->info[which].enclen;
    else
-      return iobuf->info[which].len;
+      rc = iobuf->info[which].len;
+
+   SASSERTX(rc <= sizeof(iobuf->buf[which]));
+   return rc;
 }
 
 int
@@ -548,15 +652,18 @@ socks_freeinbuffer(s, which)
         - (socks_bytesinbuffer(s, which, 0) + socks_bytesinbuffer(s, which, 1));
 
    if (sockscf.option.debug >= DEBUG_VERBOSE)
-      slog(LOG_DEBUG, "%s: socket %d, which %d, free: %lu",
-      function, s, which, (unsigned long)rc);
+      slog(LOG_DEBUG, "%s: fd %d, which %d, free: %lu",
+           function, s, which, (unsigned long)rc);
+
+   SASSERTX(rc <= sizeof(iobuf->buf[which]));
 
    return rc;
 }
 
 size_t
-socks_getfrombuffer(s, which, encoded, data, datalen)
+socks_getfrombuffer(s, flags, which, encoded, data, datalen)
    const int s;
+   const size_t flags;
    const whichbuf_t which;
    const int encoded;
    void *data;
@@ -570,16 +677,18 @@ socks_getfrombuffer(s, which, encoded, data, datalen)
       return 0;
 
    if (sockscf.option.debug >= DEBUG_VERBOSE)
-      slog(LOG_DEBUG, "%s: s = %lu, get up to %lu %s byte%s from %s buffer "
-                      "that currently has %lu decoded, %lu encoded",
-                      function,
-                      (long unsigned)s,
-                      (unsigned long)datalen,
-                      encoded ? "encoded" : "decoded",
-                      datalen == 1 ? "" : "s",
-                      which == READ_BUF ? "read" : "write",
-                      (unsigned long)socks_bytesinbuffer(s, which, 0),
-                      (unsigned long)socks_bytesinbuffer(s, which, 1));
+      slog(LOG_DEBUG,
+           "%s: fd = %d, get up to %lu %s byte%s from %s buffer which "
+           "currently has %lu decoded, %lu encoded.  Flags = %lu",
+           function,
+           s,
+           (unsigned long)datalen,
+           encoded ? "encoded" : "decoded",
+           datalen == 1 ? "" : "s",
+           which == READ_BUF ? "read" : "write",
+           (unsigned long)socks_bytesinbuffer(s, which, 0),
+           (unsigned long)socks_bytesinbuffer(s, which, 1),
+           (unsigned long)flags);
 
    if ((toget = MIN(datalen, socks_bytesinbuffer(s, which, encoded))) == 0)
       return 0;
@@ -589,23 +698,31 @@ socks_getfrombuffer(s, which, encoded, data, datalen)
 
       /* encoded data starts at the end of non-encoded data. */
       memcpy(data, &iobuf->buf[which][iobuf->info[which].len], toget);
-      iobuf->info[which].enclen -= toget;
 
-      /* move encoded data remaining in the buffer to the end of unencoded. */
-      memmove(&iobuf->buf[which][iobuf->info[which].len],
-              &iobuf->buf[which][iobuf->info[which].len + toget],
-              iobuf->info[which].enclen);
+      if (!(flags & MSG_PEEK)) {
+         iobuf->info[which].enclen -= toget;
+
+         /*
+          * encoded data stays at the end of unencoded data.
+          */
+         memmove(&iobuf->buf[which][iobuf->info[which].len],
+                 &iobuf->buf[which][iobuf->info[which].len + toget],
+                 iobuf->info[which].enclen);
+      }
    }
    else {
       SASSERTX(iobuf->info[which].len >= toget);
 
       memcpy(data, iobuf->buf[which], toget);
-      iobuf->info[which].len -= toget;
 
-      /* move the data remaining to the start of the buffer.  */
-      memmove(iobuf->buf[which],
-              &iobuf->buf[which][toget],
-              iobuf->info[which].len + iobuf->info[which].enclen);
+      if (!(flags & MSG_PEEK)) {
+         iobuf->info[which].len -= toget;
+
+         /* move the data remaining to the start of the buffer.  */
+         memmove(iobuf->buf[which],
+                 &iobuf->buf[which][toget],
+                 iobuf->info[which].len + iobuf->info[which].enclen);
+      }
    }
 
    return toget;
@@ -620,125 +737,8 @@ socks_flushallbuffers(void)
 
    for (i = 0, rc = 0; i < iobufc; ++i)
       if (iobufv[i].allocated)
-         if (socks_flushbuffer(iobufv[i].s, -1) == -1)
+         if (socks_flushbuffer(iobufv[i].s, -1, NULL) == -1)
             rc = -1;
 
    return rc;
 }
-
-#if 0
-void
-socks_iobuftest(void)
-{
-   const char *function = "socks_iobuftest()";
-   const int s = 64;
-   iobuffer_t *iobuf;
-   whichbuf_t whichv[] = { READ_BUF, WRITE_BUF };
-   size_t i, totbufsize;
-   unsigned char encbuf[MAXGSSAPITOKENLEN + GSSAPI_HLEN],
-                 decbuf[sizeof(encbuf)], tmpbuf[sizeof(encbuf)];
-
-   slog(LOG_INFO, function);
-
-   SASSERTX((iobuf = socks_allocbuffer(s, SOCK_STREAM)) != NULL);
-   SASSERTX(iobuf->s == s);
-   SASSERTX(iobuf->allocated);
-
-   slog(LOG_INFO, "%s: running simple add and retrieve test ...", function);
-   for (i = 0; i < ELEMENTS(whichv); ++i) {
-      totbufsize = socks_freeinbuffer(s, whichv[i]);
-
-      slog(LOG_INFO, "%s: free in empty buffer for %s data: %d",
-      function, i ? "encoded" : "unencoded", totbufsize);
-
-      memset(decbuf, 0xff, sizeof(decbuf));
-
-      SASSERTX(socks_addtobuffer(s, whichv[i], i, decbuf, sizeof(decbuf))
-      == sizeof(decbuf));
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], i, tmpbuf, sizeof(decbuf))
-      == sizeof(decbuf));
-
-      SASSERTX(memcmp(decbuf, tmpbuf, sizeof(tmpbuf)) == 0);
-
-      SASSERTX(socks_freeinbuffer(s, whichv[i]) == totbufsize);
-   }
-
-   slog(LOG_INFO, "%s: running mixed full add and retrieve test ...", function);
-   for (i = 0; i < ELEMENTS(whichv); ++i) {
-      totbufsize = socks_freeinbuffer(s, whichv[i]);
-
-      memset(decbuf, 0xdd, sizeof(decbuf));
-
-      SASSERTX(socks_addtobuffer(s, whichv[i], 0, decbuf, sizeof(decbuf))
-      == sizeof(decbuf));
-
-      memset(encbuf, 0xee, sizeof(encbuf));
-      SASSERTX(socks_addtobuffer(s, whichv[i], 1, encbuf, sizeof(encbuf))
-      == sizeof(decbuf));
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], 0, tmpbuf, sizeof(decbuf))
-      == sizeof(decbuf));
-      SASSERTX(memcmp(decbuf, tmpbuf, sizeof(tmpbuf)) == 0);
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], 1, tmpbuf, sizeof(encbuf))
-      == sizeof(encbuf));
-      SASSERTX(memcmp(encbuf, tmpbuf, sizeof(tmpbuf)) == 0);
-
-      SASSERTX(socks_freeinbuffer(s, whichv[i]) == totbufsize);
-   }
-
-   slog(LOG_INFO, "%s: running partial add and retrieve test ...", function);
-   for (i = 0; i < ELEMENTS(whichv); ++i) {
-      const size_t firstblock = 1024, secondblock = sizeof(tmpbuf) - firstblock;
-      totbufsize = socks_freeinbuffer(s, whichv[i]);
-
-      SASSERTX(sizeof(tmpbuf) > firstblock);
-
-      memset(decbuf, 0xd1, firstblock);
-      memset(decbuf + firstblock, 0xdf, secondblock);
-
-      memset(encbuf, 0xe1, firstblock);
-      memset(encbuf + firstblock, 0xef, secondblock);
-
-      SASSERTX(socks_addtobuffer(s, whichv[i], 0, decbuf, firstblock)
-      == firstblock);
-
-      SASSERTX(socks_addtobuffer(s, whichv[i], 1, encbuf, firstblock)
-      == firstblock);
-
-      SASSERTX(socks_addtobuffer(s, whichv[i], 0, decbuf + firstblock,
-      secondblock) == secondblock);
-
-      SASSERTX(socks_addtobuffer(s, whichv[i], 1, encbuf + firstblock,
-      secondblock) == secondblock);
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], 0, tmpbuf, firstblock)
-      == firstblock);
-      SASSERTX(memcmp(decbuf, tmpbuf, firstblock) == 0);
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], 1, tmpbuf, firstblock)
-      == firstblock);
-      SASSERTX(memcmp(encbuf, tmpbuf, firstblock) == 0);
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], 0, tmpbuf + firstblock,
-      secondblock) == secondblock);
-
-      SASSERTX(memcmp(decbuf + firstblock, tmpbuf + firstblock, secondblock)
-      == 0);
-
-      SASSERTX(socks_getfrombuffer(s, whichv[i], 1, tmpbuf + firstblock,
-      secondblock) == secondblock);
-
-      SASSERTX(memcmp(encbuf + firstblock, tmpbuf + firstblock, secondblock)
-      == 0);
-
-      SASSERTX(socks_freeinbuffer(s, whichv[i]) == totbufsize);
-      SASSERTX(socks_freeinbuffer(s, whichv[i]) == totbufsize);
-   }
-
-   socks_freebuffer(s);
-   SASSERTX(iobuf->s == 0);
-   SASSERTX(!iobuf->allocated);
-}
-#endif
