@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009, 2010, 2011
+ * Copyright (c) 2009, 2010, 2011, 2012, 2013
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -51,66 +51,53 @@
 #if HAVE_GSSAPI
 
 static const char rcsid[] =
-   "$Id: method_gssapi.c,v 1.55 2011/06/27 13:18:05 michaels Exp $";
+   "$Id: method_gssapi.c,v 1.73 2013/10/27 15:24:42 karls Exp $";
 
 static negotiate_result_t
-recv_gssapi_auth_ver(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_auth_ver(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_auth_type(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_auth_type(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_auth_len(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_auth_len(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_auth_token(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_auth_token(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_enc_ver(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_enc_ver(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_enc_type(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_enc_type(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_enc_len(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_enc_len(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_enc_token(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_enc_token(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_packet(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_packet(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_packet_ver(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_packet_ver(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_packet_type(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_packet_type(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_packet_len(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_packet_len(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_gssapi_packet_token(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_gssapi_packet_token(int s, request_t *request, negotiate_state_t *state);
 
 
 negotiate_result_t
 method_gssapi(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 
 {
    const char *function = "method_gssapi()";
@@ -137,8 +124,8 @@ method_gssapi(s, request, state)
 static negotiate_result_t
 recv_gssapi_auth_ver(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_auth_ver()";
    unsigned char gssapi_auth_version;
@@ -152,8 +139,10 @@ recv_gssapi_auth_ver(s, request, state)
 
       default:
          snprintf(state->emsg, sizeof(state->emsg),
-         "%s: unknown version on gssapi packet from client: %d",
-         function, gssapi_auth_version);
+                  "%s: unknown %s version on gssapi packet from client: %d",
+                  function,
+                  proxyprotocol2string(request->version),
+                  gssapi_auth_version);
 
          return NEGOTIATE_ERROR;
    }
@@ -162,12 +151,11 @@ recv_gssapi_auth_ver(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-
 static negotiate_result_t
 recv_gssapi_auth_type(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_auth_type()";
    unsigned char gssapi_auth_type;
@@ -203,8 +191,8 @@ recv_gssapi_auth_type(s, request, state)
 static negotiate_result_t
 recv_gssapi_auth_len(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    INIT(sizeof(state->gssapitoken_len));
    CHECK(&state->gssapitoken_len, request->auth, NULL);
@@ -219,10 +207,11 @@ recv_gssapi_auth_len(s, request, state)
 static negotiate_result_t
 recv_gssapi_auth_token(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_auth_token()";
+   sendto_info_t sendtoflags;
    gss_name_t            client_name   = GSS_C_NO_NAME;
    gss_name_t            server_name   = GSS_C_NO_NAME;
    gss_cred_id_t         server_creds  = GSS_C_NO_CREDENTIAL;
@@ -243,8 +232,7 @@ recv_gssapi_auth_token(s, request, state)
 
    CHECK(input_token.value, request->auth, NULL);
 
-   SASSERTX(strlen(request->auth->mdata.gssapi.keytab) < sizeof(env));
-   strcpy(env, request->auth->mdata.gssapi.keytab);
+   STRCPY_ASSERTLEN(env, request->auth->mdata.gssapi.keytab);
    setenv("KRB5_KTNAME", env, 1);
 
 #if HAVE_HEIMDAL_KERBEROS
@@ -314,11 +302,11 @@ recv_gssapi_auth_token(s, request, state)
    sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_OFF);
 
    slog(LOG_DEBUG, "%s: length of output_token is %lu",
-   function, (unsigned long)output_token.length);
+        function, (unsigned long)output_token.length);
 
    if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
       snprintf(state->emsg, sizeof(state->emsg),
-      "%s: gss_accept_sec_context() failed: %s", function, emsg);
+               "%s: gss_accept_sec_context() failed: %s", function, emsg);
 
       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
       return NEGOTIATE_ERROR;
@@ -355,8 +343,18 @@ recv_gssapi_auth_token(s, request, state)
 
    CLEAN_GSS_TOKEN(output_token);
 
-   if ((rc = socks_sendton(s, buf, buflen, 0, 0, NULL, 0, request->auth))
-   != (ssize_t)buflen)  {
+   bzero(&sendtoflags, sizeof(sendtoflags));
+   sendtoflags.side = INTERNALIF;
+
+   if ((rc = socks_sendton(s,
+                           buf,
+                           buflen,
+                           0,
+                           0,
+                           NULL,
+                           0,
+                           NULL,
+                           request->auth)) != (ssize_t)buflen)  {
       snprintf(state->emsg, sizeof(state->emsg),
               "socks_sendton() token: wrote %ld out of %lu byte%s: %s",
                (long)rc,
@@ -375,7 +373,7 @@ recv_gssapi_auth_token(s, request, state)
 
       if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
          snprintf(state->emsg, sizeof(state->emsg),
-         "%s: gss_display_name(): %s", function, emsg);
+                  "%s: gss_display_name(): %s", function, emsg);
 
          CLEAN_GSS_AUTH(client_name, server_name, server_creds);
          return NEGOTIATE_ERROR;
@@ -396,11 +394,13 @@ recv_gssapi_auth_token(s, request, state)
 
       /* expect a new token, with the version, length, etc. header. */
       state->rcurrent = recv_gssapi_auth_ver;
-      return state->rcurrent(s, request, state);
+
+      /* presumably client is awaiting our response. */
+      return NEGOTIATE_CONTINUE;
    }
    else {
       snprintf(state->emsg, sizeof(state->emsg),
-      "%s: unknown gss major_status %d", function, major_status);
+               "%s: unknown gss major_status %d", function, major_status);
 
       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
       return NEGOTIATE_ERROR;
@@ -412,8 +412,8 @@ recv_gssapi_auth_token(s, request, state)
 static negotiate_result_t
 recv_gssapi_enc_ver(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_enc_ver()";
    unsigned char gssapi_enc_version;
@@ -450,8 +450,8 @@ recv_gssapi_enc_ver(s, request, state)
 static negotiate_result_t
 recv_gssapi_enc_type(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_enc_type()";
    unsigned char gssapi_enc_type;
@@ -486,8 +486,8 @@ recv_gssapi_enc_type(s, request, state)
 static negotiate_result_t
 recv_gssapi_enc_len(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    INIT(sizeof(state->gssapitoken_len));
@@ -503,18 +503,18 @@ recv_gssapi_enc_len(s, request, state)
 static negotiate_result_t
 recv_gssapi_enc_token(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_enc_token()";
    OM_uint32  minor_status, major_status;
-   gss_buffer_desc          input_token, output_token = GSS_C_EMPTY_BUFFER;
-   unsigned short           token_length;
-   ssize_t                  rc;
-   size_t                   buflen;
-   int                      gss_enc, gss_server_enc, conf_state, rfc1961;
-   char                     emsg[1024];
-   unsigned char            buf[GSSAPI_HLEN + MAXGSSAPITOKENLEN], c;
+   gss_buffer_desc input_token, output_token = GSS_C_EMPTY_BUFFER;
+   unsigned short  unsigned_short;
+   ssize_t rc;
+   size_t buflen;
+   unsigned char buf[GSSAPI_HLEN + MAXGSSAPITOKENLEN], gss_enc, gss_server_enc;
+   int conf_state, do_rfc1961;
+   char emsg[1024];
 
    INIT(state->gssapitoken_len);
 
@@ -523,28 +523,30 @@ recv_gssapi_enc_token(s, request, state)
 
    CHECK(input_token.value, request->auth, NULL);
 
-   /* If token length = 1 => clear text encryption selection packet */
-   if (state->gssapitoken_len == 1
+   if (state->gssapitoken_len == 1 /* clear text encryption selection packet */
    &&  request->auth->mdata.gssapi.encryption.nec)
-      rfc1961 = 0; /*
-                    * It seems the NEC reference
-                    * implementation does not do this right.
-                    */
+      do_rfc1961 = 0; /*
+                       * It seems the NEC reference implementation does not do
+                       * this right.  Try to support it if configured as such.
+                       */
    else
-      rfc1961 = 1;
+      do_rfc1961 = 1;
 
-   if (rfc1961) {
-      slog(LOG_DEBUG,
-      "%s: rule assumes client uses rfc1961 encrypted exchange", function);
+   slog(LOG_DEBUG, "%s: rule assumes client uses %s exchange",
+                   function,
+                   do_rfc1961? "rfc1961 encrypted" : "nec unencrypted");
 
-      major_status
-      = gss_unwrap(&minor_status,
-                   request->auth->mdata.gssapi.state.id,
-                   &input_token, &output_token, 0, GSS_C_QOP_DEFAULT);
+   if (do_rfc1961) {
+      major_status = gss_unwrap(&minor_status,
+                                request->auth->mdata.gssapi.state.id,
+                                &input_token,
+                                &output_token,
+                                0,
+                                GSS_C_QOP_DEFAULT);
 
       if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
-         snprintf(state->emsg, sizeof(state->emsg),
-         "%s: gss_unwrap(): %s", function, emsg);
+         snprintf(state->emsg, sizeof(state->emsg), "%s: gss_unwrap(): %s",
+                  function, emsg);
 
          CLEAN_GSS_TOKEN(output_token);
          return NEGOTIATE_ERROR;
@@ -552,30 +554,25 @@ recv_gssapi_enc_token(s, request, state)
 
       if (output_token.length != 1) {
          snprintf(state->emsg, sizeof(state->emsg),
-                  "%s: gssapi encryption unwrapped length is wrong: is %lu, "
-                  "but expected is 1.  Is the client a NEC-based client "
-                  "perhaps?  If so, try enabling NEC client-compatibility for "
-                  "this client-rule",
+                  "%s: gssapi encryption unwrapped length is wrong: expected 1 "
+                  "but is %lu.  If the client is a NEC-based client you can "
+                  "try enabling NEC (NOTE: unencrypted exchange) "
+                  "client-compatibility for this client-rule",
                   function, (unsigned long)output_token.length);
 
          CLEAN_GSS_TOKEN(output_token);
          return NEGOTIATE_ERROR;
       }
 
-      SASSERTX(output_token.length == sizeof(c));
-      memcpy(&c, output_token.value, output_token.length);
+      SASSERTX(output_token.length == sizeof(gss_enc));
+      memcpy(&gss_enc, output_token.value, output_token.length);
       CLEAN_GSS_TOKEN(output_token);
    }
    else {
-      slog(LOG_DEBUG,
-      "%s: rule assumes client uses nec unencrpypted exchange", function);
-
-      memcpy(&c, input_token.value, 1);
+      SASSERTX(input_token.length == sizeof(gss_enc));
+      memcpy(&gss_enc, input_token.value, input_token.length);
    }
 
-   gss_enc = (int)c;
-   token_length = htons(1);
-
    if ((gss_enc == SOCKS_GSSAPI_CLEAR
     && !request->auth->mdata.gssapi.encryption.clear)
    ||  (gss_enc == SOCKS_GSSAPI_INTEGRITY
@@ -588,9 +585,9 @@ recv_gssapi_enc_token(s, request, state)
        */
 
       snprintf(state->emsg, sizeof(state->emsg),
-      "the client requests different authentication from what we offer.\n"
-      "Client requests: %s, "
-      "We offer: clear/%d, integrity/%d, confidentiality/%d, per message/%d",
+      "the client requests different authentication from what we offer.  "
+      "Client requests %s, but we offer: clear/%d, integrity/%d, "
+      "confidentiality/%d, per message/%d",
       gssapiprotection2string(gss_enc),
       request->auth->mdata.gssapi.encryption.clear,
       request->auth->mdata.gssapi.encryption.integrity,
@@ -602,10 +599,10 @@ recv_gssapi_enc_token(s, request, state)
 
    gss_server_enc = gss_enc;
 
-   slog(LOG_DEBUG, "%s: gssapi: using %s protection",
-   function, gssapiprotection2string(gss_server_enc));
+   slog(LOG_DEBUG, "%s: using %s protection (%d)",
+        function, gssapiprotection2string(gss_server_enc), gss_server_enc);
 
-   if (rfc1961) {
+   if (do_rfc1961) {
       input_token.length = 1;
       input_token.value  = &gss_server_enc;
 
@@ -618,16 +615,17 @@ recv_gssapi_enc_token(s, request, state)
                               &output_token);
 
       if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
-         snprintf(state->emsg, sizeof(state->emsg),
-         "%s: gss_wrap(): %s", function, emsg);
+         snprintf(state->emsg, sizeof(state->emsg), "%s: gss_wrap(): %s",
+                  function, emsg);
 
          CLEAN_GSS_TOKEN(output_token);
          return NEGOTIATE_ERROR;
       }
-
-      token_length = htons(output_token.length);
    }
-
+   else {
+      output_token.length  = 1;
+      output_token.value   = &gss_server_enc;
+   }
 
    request->auth->mdata.gssapi.state.protection = gss_server_enc;
 
@@ -642,35 +640,35 @@ recv_gssapi_enc_token(s, request, state)
     *
     */
 
-   SASSERTX((size_t)GSSAPI_HLEN + ntohs(token_length) <= sizeof(buf));
+   SASSERTX(GSSAPI_HLEN + output_token.length <= sizeof(buf));
 
    buflen = 0;
    buf[buflen++] = SOCKS_GSSAPI_VERSION;
    buf[buflen++] = SOCKS_GSSAPI_ENCRYPTION;
 
-   memcpy(&buf[buflen], &token_length, sizeof(token_length));
-   buflen += sizeof(token_length);
+   unsigned_short = htons(output_token.length);
+   memcpy(&buf[buflen], &unsigned_short, sizeof(unsigned_short));
+   buflen += sizeof(unsigned_short);
 
-   if (rfc1961)
-      memcpy(&buf[buflen], output_token.value, ntohs(token_length));
-   else  /* send unprotected reply */
-      memcpy(&buf[buflen], &gss_server_enc, ntohs(token_length));
-   buflen += ntohs(token_length);
+   memcpy(&buf[buflen], output_token.value, output_token.length);
+   buflen += (output_token.length);
 
-   if ((rc = socks_sendton(s, buf, buflen, 0, 0, NULL, 0, request->auth))
+   if ((rc = socks_sendton(s, buf, buflen, 0, 0, NULL, 0, NULL, request->auth))
    != (ssize_t)buflen) {
       snprintf(state->emsg, sizeof(state->emsg),
-               "socks_sendton() buf: wrote %ld out of %lu byte%s: %s",
+               "socks_sendton() buf: wrote %ld out of %lu bytes: %s",
                 (long)rc,
                 (unsigned long)buflen,
-                buflen == 1 ? "" : "s",
                 strerror(errno));
 
-      CLEAN_GSS_TOKEN(output_token);
+      if (do_rfc1961)
+         CLEAN_GSS_TOKEN(output_token);
+
       return NEGOTIATE_ERROR;
    }
 
-   CLEAN_GSS_TOKEN(output_token);
+   if (do_rfc1961)
+      CLEAN_GSS_TOKEN(output_token);
 
    if (gss_server_enc)
       state->rcurrent = recv_gssapi_packet;
@@ -681,7 +679,7 @@ recv_gssapi_enc_token(s, request, state)
        */
       state->rcurrent = recv_sockspacket;
 
-   return state->rcurrent(s, request, state);
+   return NEGOTIATE_CONTINUE; /* presumably client is awaiting our response. */
 }
 
 /*
@@ -697,8 +695,8 @@ recv_gssapi_enc_token(s, request, state)
 static negotiate_result_t
 recv_gssapi_packet(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 
 {
 
@@ -709,8 +707,8 @@ recv_gssapi_packet(s, request, state)
 static negotiate_result_t
 recv_gssapi_packet_ver(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_packet_ver()";
    unsigned char gssapi_packet_version;
@@ -737,8 +735,8 @@ recv_gssapi_packet_ver(s, request, state)
 static negotiate_result_t
 recv_gssapi_packet_type(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_packet_type()";
    unsigned char gssapi_enc_type;
@@ -773,8 +771,8 @@ recv_gssapi_packet_type(s, request, state)
 static negotiate_result_t
 recv_gssapi_packet_len(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    INIT(sizeof(state->gssapitoken_len));
@@ -790,8 +788,8 @@ recv_gssapi_packet_len(s, request, state)
 static negotiate_result_t
 recv_gssapi_packet_token(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_gssapi_packet_token()";
    OM_uint32        minor_status, major_status      = GSS_S_COMPLETE;
@@ -825,7 +823,7 @@ recv_gssapi_packet_token(s, request, state)
    }
 
    offset = 0;
-   data = output_token.value;
+   data   = output_token.value;
 
    if (offset + sizeof(request->version) > output_token.length) {
       snprintf(state->emsg, sizeof(state->emsg),
@@ -835,6 +833,7 @@ recv_gssapi_packet_token(s, request, state)
       CLEAN_GSS_TOKEN(output_token);
       return NEGOTIATE_ERROR;
    }
+
    memcpy(&request->version, &data[offset], sizeof(request->version));
    if (request->version != PROXY_SOCKS_V5) {
       snprintf(state->emsg, sizeof(state->emsg),
