@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2005, 2008, 2009, 2010,
- *               2011
+ *               2011, 2012, 2013
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -45,73 +45,79 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: sockd_protocol.c,v 1.167 2011/07/08 10:11:52 michaels Exp $";
+"$Id: sockd_protocol.c,v 1.214 2013/10/27 15:24:42 karls Exp $";
 
 #if SOCKS_SERVER
 static negotiate_result_t
-recv_v4req(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_v4req(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_v5req(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_v5req(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_methods(int s, struct request_t *request,
-             struct negotiate_state_t *state);
+recv_methods(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_ver(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_ver(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_cmd(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_cmd(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_flag(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_flag(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_sockshost(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+recv_sockshost(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_atyp(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_atyp(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_port(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_port(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_address(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_address(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_domain(int s, struct request_t *request, struct negotiate_state_t *state);
+recv_domain(int s, request_t *request, negotiate_state_t *state);
 
 static negotiate_result_t
-recv_username(int s, struct request_t *request,
-              struct negotiate_state_t *state);
+recv_username(int s, request_t *request,
+              negotiate_state_t *state);
 
 static negotiate_result_t
-methodnegotiate(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+methodnegotiate(int s, request_t *request,
+      negotiate_state_t *state);
 #endif /* SOCKS_SERVER */
 
 negotiate_result_t
 recv_clientrequest(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_clientrequest()";
 #if HAVE_NEGOTIATE_PHASE
    negotiate_result_t rc;
 #endif /* HAVE_NEGOTIATE_PHASE */
 
-   slog(LOG_DEBUG, "%s: client %s, read so far: %lu",
-                   function,
-                   sockshost2string(&state->src, NULL, 0),
-                   (unsigned long)state->reqread);
+   slog(LOG_DEBUG,
+        "%s: fd %d, client %s, state->complete: %d, read so far: %lu",
+        function,
+        s,
+        sockshost2string(&state->src, NULL, 0),
+        state->complete,
+        (unsigned long)state->reqread);
 
    if (state->complete)
       return NEGOTIATE_FINISHED;
 
 #if HAVE_NEGOTIATE_PHASE
+
+#if SOCKS_SERVER
+   CTASSERT(sizeof(state->mem) > (MAXMETHODS + MAXNAMELEN + MAXPWLEN));
+#endif /* SOCKS_SERVER */
+
    if (state->rcurrent != NULL)   /* not first call on this client. */
       rc = state->rcurrent(s, request, state);
    else {
@@ -132,7 +138,7 @@ recv_clientrequest(s, request, state)
 
          default:
             snprintf(state->emsg, sizeof(state->emsg),
-                     "unknown socks version %d in clientrequest",
+                     "unknown SOCKS version %d in client request",
                      request->version);
             return NEGOTIATE_ERROR;
       }
@@ -149,7 +155,6 @@ recv_clientrequest(s, request, state)
                       sockshost2string(&state->src, src, sizeof(src)),
                       sockshost2string(&state->dst, dst, sizeof(dst)));
 
-
       rc = state->rcurrent(s, request, state);
    }
 
@@ -158,6 +163,7 @@ recv_clientrequest(s, request, state)
 #else /* !HAVE_NEGOTIATE_PHASE */
 
    SASSERTX(state->complete);
+   return NEGOTIATE_FINISHED;
    /* NOTREACHED */
 
 #endif /* !HAVE_NEGOTIATE_PHASE */
@@ -166,27 +172,35 @@ recv_clientrequest(s, request, state)
 #if HAVE_NEGOTIATE_PHASE
 void
 send_failure(s, response, failure)
-   int s;
-   struct response_t *response;
-   int failure;
+   const int s;
+   response_t *response;
+   const unsigned int failure;
 {
-#if HAVE_GSSAPI
    const char *function = "send_failure()";
-   gss_buffer_desc output_token;
-   OM_uint32 minor_status;
-#endif /* HAVE_GSSAPI */
 
    socks_set_responsevalue(response, sockscode(response->version, failure));
-   send_response(s, response);
+   if (send_response(s, response) != 0)
+      slog(LOG_DEBUG, "%s: could not send failure to client: %s",
+           function, strerror(errno));
 
 #if HAVE_GSSAPI
-   if (response->auth->method == AUTHMETHOD_GSSAPI) {
-      if (gss_delete_sec_context(&minor_status,
-         &response->auth->mdata.gssapi.state.id, &output_token)
-         != GSS_S_COMPLETE)
-            swarn("%s: gss_delete_sec_context failed", function);
+   if (response->auth->method == AUTHMETHOD_GSSAPI
+   && response->auth->mdata.gssapi.state.id != GSS_C_NO_CONTEXT) {
+      OM_uint32 major_status, minor_status;
+      char buf[512];
 
-      CLEAN_GSS_TOKEN(output_token);
+      if ((major_status
+      = gss_delete_sec_context(&minor_status,
+                               &response->auth->mdata.gssapi.state.id,
+                               GSS_C_NO_BUFFER)) != GSS_S_COMPLETE) {
+         if (!gss_err_isset(major_status, minor_status, buf, sizeof(buf)))
+            *buf = NUL;
+
+         swarn("%s: gss_delete_sec_context() failed%s%s",
+               function,
+               *buf == NUL ? "" : ": ",
+               *buf == NUL ? "" : buf);
+      }
    }
 #endif /* HAVE_GSSAPI */
 }
@@ -194,14 +208,16 @@ send_failure(s, response, failure)
 int
 send_response(s, response)
    int s;
-   const struct response_t *response;
+   const response_t *response;
 {
    const char *function = "send_response()";
-   ssize_t sendt;
+   iobuffer_t *tmpiobuf;
+   sendto_info_t sendtoflags;
+   ssize_t sent;
    size_t length;
 #if SOCKS_SERVER
    unsigned char responsemem[sizeof(*response)], *p = responsemem;
-#else
+#else /* !SOCKS_SERVER */
    char responsemem[1024], *p = responsemem;
 #endif
 
@@ -280,16 +296,41 @@ send_response(s, response)
    length = p - responsemem;
 
 #if SOCKS_SERVER
-   slog(LOG_DEBUG, "%s: sending response: %s",
-   function, socks_packet2string(response, 0));
+   slog(LOG_DEBUG, "%s: sending response: %s, authmethod %d",
+        function, socks_packet2string(response, 0), response->auth->method);
+
 #else /* COVENANT */
    slog(LOG_DEBUG, "%s: sending response:\n%s", function, responsemem);
 #endif
 
-   if ((sendt = socks_sendton(s, responsemem, length, 1, 0, NULL, 0,
-   response->auth)) != (ssize_t)length) {
+   /*
+    * If sending response from a process that normally does not send
+    * any response, and thus does not allocate a buffer for this fd.
+    */
+   if (socks_getbuffer(s) == NULL)
+      tmpiobuf = socks_allocbuffer(s, SOCK_STREAM);
+   else
+      tmpiobuf = NULL;
+
+   bzero(&sendtoflags, sizeof(sendtoflags));
+   sendtoflags.side = INTERNALIF;
+
+   sent = socks_sendton(s,
+                        responsemem,
+                        length,
+                        1,
+                        0,
+                        NULL,
+                        0,
+                        &sendtoflags,
+                        response->auth);
+
+   if (tmpiobuf != NULL)
+      socks_freebuffer(s);
+
+   if (sent  != (ssize_t)length) {
       slog(LOG_DEBUG, "%s: socks_sendton(): %ld/%lu: %s",
-      function, (long)sendt, (long unsigned)length, strerror(errno));
+           function, (long)sent, (long unsigned)length, strerror(errno));
 
       return -1;
    }
@@ -297,12 +338,85 @@ send_response(s, response)
    return 0;
 }
 
+int
+send_connectresponse(s, error, io)
+   const int s;
+   const int error;
+   sockd_io_t *io;
+{
+   response_t response;
+
+   create_response(
+#if SOCKS_SERVER
+                   sockaddr2sockshost(&io->dst.laddr, NULL),
+#elif COVENANT
+                   &io->dst.host,
+#else /* !COVENANT */
+#error "who are we?"
+#endif
+
+                   &io->src.auth,
+                   io->state.proxyprotocol,
+                   error == 0 ?
+                     SOCKS_SUCCESS : (int)errno2reply(error,
+                                                      io->state.proxyprotocol),
+                   &response);
+
+   if (send_response(s, &response) != 0)
+      return -1;
+
+   return 0;
+}
+
+response_t *
+create_response(host, auth, version, responsecode, response)
+   const sockshost_t *host;
+   authmethod_t *auth;
+   const int version;
+   const int responsecode;
+   response_t *response;
+{
+
+   bzero(response, sizeof(*response));
+   response->auth = auth;
+
+   switch (version) {
+#if SOCKS_SERVER
+      case PROXY_SOCKS_V4:
+         response->version = PROXY_SOCKS_V4REPLY_VERSION;
+         break;
+
+      case PROXY_SOCKS_V5:
+#elif COVENANT /* !SOCKS_SERVER */
+      case PROXY_HTTP_10:
+      case PROXY_HTTP_11:
+#endif /* COVENANT */
+         response->version = version;
+         break;
+
+      default:
+         SERRX(version);
+   }
+
+   if (host != NULL)
+      response->host = *host;
+   else
+      response->host.atype = SOCKS_ADDR_IPV4;
+      /* rest can be 0. */
+
+   socks_set_responsevalue(response,
+                           sockscode(version, responsecode));
+
+   return response;
+}
+
+
 #if SOCKS_SERVER
 negotiate_result_t
 recv_sockspacket(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    return recv_ver(s, request, state);
@@ -311,8 +425,8 @@ recv_sockspacket(s, request, state)
 static negotiate_result_t
 recv_v4req (s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    /*
@@ -323,6 +437,11 @@ recv_v4req (s, request, state)
     * so minimum length is 9.
     */
 
+   /*
+    * No methods supported in v4.
+    */
+   request->auth->method = AUTHMETHOD_NONE;
+
    /* CD */
    state->rcurrent = recv_cmd;
    return state->rcurrent(s, request, state);
@@ -332,8 +451,8 @@ recv_v4req (s, request, state)
 static negotiate_result_t
 recv_v5req (s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    /*
@@ -377,39 +496,60 @@ recv_v5req (s, request, state)
 static negotiate_result_t
 recv_methods(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 
 {
    const char *function = "recv_methods()";
+   sendto_info_t sendtoflags;
    const unsigned char methodc = state->mem[AUTH_NMETHODS];
-   unsigned char reply[ 1 /* VERSION   */
-                      + 1 /* METHOD    */
+   unsigned char reply[   1 /* VERSION   */
+                        + 1 /* METHOD    */
                       ];
-   char buf[256 * (sizeof("0x00") + sizeof(", ")) + 1];
+   char buf[(AUTHMETHOD_MAX + 1) * (sizeof("0x00 (some methodname")
+            + sizeof(", ")) + 1];
    size_t bufused;
    int i;
 
    INIT(methodc);
    CHECK(&state->mem[start], request->auth, NULL);
 
+   *buf = NUL;
    for (i = bufused = 0; i < methodc; ++i)
-      bufused += snprintf(&buf[bufused], sizeof(buf) - bufused, "0x%x, ",
-      state->mem[start + i]);
+      bufused += snprintf(&buf[bufused], sizeof(buf) - bufused, "0x%x (%s), ",
+                          state->mem[start + i],
+                          method2string(state->mem[start + i]));
 
-   STRIPTRAILING(buf, bufused, ", \t\n");
+   if (bufused >= sizeof(buf)  - 1)
+      swarnx("%s: suspiciously many (%u) methods offered by client %s",
+             function, methodc, sockshost2string(&state->src, NULL, 0));
 
-   slog(LOG_DEBUG, "%s: client offered %d authentication method%s: %s",
-   function, methodc, methodc == 1 ? "" : "s", buf);
+   if (bufused >= strlen(", ") && buf[bufused - strlen(", ")] == ',')
+      buf[bufused - strlen(", ")] = NUL;
+
+   slog(LOG_DEBUG, "%s: client %s offered %d authentication method%s: %s",
+        function,
+        sockshost2string(&state->src, NULL, 0),
+        methodc, methodc == 1 ? "" : "s",
+        buf);
 
    switch (request->auth->method) {
       case AUTHMETHOD_NOTSET:
-         slog(LOG_DEBUG, "%s: method set to %d, selecting new",
-         function, request->auth->method);
-
-         request->auth->method = selectmethod(sockscf.methodv, sockscf.methodc,
-         &state->mem[start], (size_t)methodc);
+         slog(LOG_DEBUG,
+              "%s: socksmethod to use not set, selecting amongst the "
+              "following %lu method%s: %s",
+              function,
+              (unsigned long)state->crule->state.smethodc,
+              (unsigned long)state->crule->state.smethodc == 1 ? "" : "s",
+              methods2string(state->crule->state.smethodc,
+                             state->crule->state.smethodv,
+                             NULL,
+                             0));
 
+         request->auth->method = selectmethod(state->crule->state.smethodv,
+                                              state->crule->state.smethodc,
+                                              &state->mem[start],
+                                              (size_t)methodc);
          break;
 
       default: {
@@ -422,9 +562,9 @@ recv_methods(s, request, state)
           */
          size_t i;
 
-         slog(LOG_DEBUG, "%s: method %d already chosen for this rule, "
-                         "not selecting again",
-                         function, request->auth->method);
+         slog(LOG_DEBUG,
+              "%s: method %d already chosen for this rule, not selecting again",
+              function, request->auth->method);
 
          for (i = 0; i < methodc; ++i)
             if (state->mem[start + i] == request->auth->method)
@@ -446,32 +586,45 @@ recv_methods(s, request, state)
     *   +----+--------+
     */
 
-   slog(LOG_DEBUG, "%s: sending authentication reply: VER: %d METHOD: %d",
-   function, request->version, request->auth->method);
+   slog(LOG_DEBUG, "%s: sending authentication reply: VER: %d METHOD: %d (%s)",
+        function,
+        request->version,
+        request->auth->method,
+        method2string(request->auth->method));
 
-   reply[AUTH_VERSION]  = request->version;
-   reply[AUTH_METHOD]   = (unsigned char)request->auth->method;
+   reply[AUTH_VERSION]        = request->version;
+   reply[AUTH_SELECTEDMETHOD] = (unsigned char)request->auth->method;
 
-   if (socks_sendton(s, reply, sizeof(reply), sizeof(reply), 0, NULL,
-   0, request->auth) != sizeof(reply))
+   bzero(&sendtoflags, sizeof(sendtoflags));
+   sendtoflags.side = EXTERNALIF;
+
+   if (socks_sendton(s,
+                     reply,
+                     sizeof(reply),
+                     sizeof(reply),
+                     0,
+                     NULL,
+                     0,
+                     &sendtoflags,
+                     request->auth) != sizeof(reply))
       return NEGOTIATE_ERROR;
 
    if (request->auth->method == AUTHMETHOD_NOACCEPT) {
       snprintf(state->emsg, sizeof(state->emsg),
-      "client offered no acceptable authentication method");
+              "client offered no acceptable authentication method");
 
       return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = methodnegotiate;
-   return state->rcurrent(s, request, state);
+   return NEGOTIATE_CONTINUE; /* presumably client is awaiting our response. */
 }
 
 static negotiate_result_t
 methodnegotiate(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    /* authentication method dependent negotiation */
@@ -500,8 +653,8 @@ methodnegotiate(s, request, state)
 static negotiate_result_t
 recv_ver(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    /* VER */
@@ -527,8 +680,8 @@ recv_ver(s, request, state)
 static negotiate_result_t
 recv_cmd(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_cmd()";
 
@@ -546,8 +699,12 @@ recv_cmd(s, request, state)
          break;
 
       default:
-         /*  XXX should print to emsg, not swarn() here. */
-         swarnx("%s: unknown command: %d", function, request->command);
+         snprintf(state->emsg, sizeof(state->emsg),
+                  "unknown %s command: %d",
+                  proxyprotocol2string(request->version),
+                  request->command);
+
+         return NEGOTIATE_ERROR;
    }
 
    switch (request->version) {
@@ -569,8 +726,8 @@ recv_cmd(s, request, state)
 static negotiate_result_t
 recv_flag(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    INIT(sizeof(request->flag));
@@ -582,8 +739,8 @@ recv_flag(s, request, state)
 static negotiate_result_t
 recv_sockshost(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    switch (request->version) {
       case PROXY_SOCKS_V4:
@@ -604,8 +761,8 @@ recv_sockshost(s, request, state)
 static negotiate_result_t
 recv_atyp(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    INIT(sizeof(request->host.atype));
@@ -617,8 +774,8 @@ recv_atyp(s, request, state)
 static negotiate_result_t
 recv_address(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    switch (request->version) {
@@ -626,7 +783,7 @@ recv_address(s, request, state)
          INIT(sizeof(request->host.addr.ipv4));
 
          /* only one supported in v4. */
-         request->host.atype = (unsigned char)SOCKS_ADDR_IPV4;
+         request->host.atype = SOCKS_ADDR_IPV4;
 
          CHECK(&request->host.addr.ipv4, request->auth, recv_username);
          SERRX(0); /* NOTREACHED */
@@ -641,8 +798,8 @@ recv_address(s, request, state)
             }
 
             case SOCKS_ADDR_IPV6: {
-               INIT(sizeof(request->host.addr.ipv6));
-               CHECK(&request->host.addr.ipv6, request->auth, recv_port);
+               INIT(sizeof(request->host.addr.ipv6.ip));
+               CHECK(&request->host.addr.ipv6.ip, request->auth, recv_port);
                SERRX(0); /* NOTREACHED */
             }
 
@@ -658,8 +815,10 @@ recv_address(s, request, state)
             }
 
             default:
-               slog(LOG_DEBUG, "unknown address format %d in reply",
-               request->host.atype);
+               snprintf(state->emsg, sizeof(state->emsg),
+                        "unknown %s command: %d",
+                        proxyprotocol2string(request->version),
+                        request->command);
                return NEGOTIATE_ERROR;
          }
 
@@ -673,8 +832,8 @@ recv_address(s, request, state)
 static negotiate_result_t
 recv_domain(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    unsigned char alen;
    /* first byte gives length. */
@@ -695,8 +854,8 @@ recv_domain(s, request, state)
 static negotiate_result_t
 recv_port(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
 
    INIT(sizeof(request->host.port));
@@ -720,8 +879,8 @@ recv_port(s, request, state)
 static negotiate_result_t
 recv_username(s, request, state)
    int s;
-   struct request_t *request;
-   struct negotiate_state_t *state;
+   request_t *request;
+   negotiate_state_t *state;
 {
    const char *function = "recv_username()";
    char *username = (char *)&state->mem[sizeof(request->version)
@@ -744,9 +903,20 @@ recv_username(s, request, state)
 
          state->mem[state->reqread - 1] = NUL;
 
-         swarnx("%s: username too long (> %ld): \"%s...\"", function,
-         (unsigned long)strlen(username),
-         str2vis(username, strlen(username), visstring, sizeof(visstring)));
+         snprintf(state->emsg, sizeof(state->emsg),
+                  "%s username received from client is too long.  Max length "
+                  "s %lu, length of received name is longer than %lu: "
+                  "\"%s ...\"",
+                  proxyprotocol2string(request->version),
+                  (unsigned long)(MAXNAMELEN - 1),
+                  (unsigned long)strlen(username),
+                  str2vis(username,
+                          strlen(username),
+                          visstring,
+                          sizeof(visstring)));
+
+         slog(LOG_NOTICE, "%s: strange data from client %s: %s",
+              function, sockshost2string(&state->src, NULL, 0), state->emsg);
 
          return NEGOTIATE_ERROR;
       }
